/*
Copyright 2023 ycyxuehan.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package v1alpha1

import (
	"path"
	"strings"

	"github.com/ycyxuehan/zelda/pkg/utils"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
)

type Phase string

const (
	PhaseCreating  Phase  = "Creating"
	PhaseInitial   Phase  = "Initial"
	PhaseBuild     Phase  = "Build"
	PhasePackage   Phase  = "Package"
	PhaseDeploy    Phase  = "Deploy"
	PhaseCompleted Phase  = "Completed"
	PhasePending   Phase  = "Pending"
	Succeeded      string = "Succeeded"
)

// EDIT THIS FILE!  THIS IS SCAFFOLDING FOR YOU TO OWN!
// NOTE: json tags are required.  Any new fields you add must have json tags for the fields to be serialized.
//Step status
type StepStatus struct {
	Name       string          `json:"name,omitempty"`
	Phase      corev1.PodPhase `json:"phase,omitempty"`
	StartedAt  metav1.Time     `json:"startedAt,omitempty"`
	FinishedAt metav1.Time     `json:"finishedAt,omitempty"`
	Started    *bool           `json:"started,omitempty"`
	Ready      bool            `json:"ready,omitempty"`
	Message    string          `json:"message,omitempty"`
	Reason     string          `json:"reason,omitempty"`
	// corev1.ContainerStatus
}

func NewStepStatus(status corev1.ContainerStatus) StepStatus {
	stepStatus := StepStatus{}
	stepStatus.Started = status.Started
	stepStatus.Ready = status.Ready
	stepStatus.Name = status.Name
	//等待
	if status.State.Waiting != nil {
		stepStatus.Phase = corev1.PodPending
		stepStatus.Message = status.State.Waiting.Message
		stepStatus.Reason = status.State.Waiting.Reason
	}
	//运行中
	if status.State.Running != nil {
		stepStatus.Phase = corev1.PodRunning
		stepStatus.StartedAt = status.State.Running.StartedAt
	}
	if status.State.Terminated != nil {
		if status.State.Terminated.ExitCode != 0 {
			stepStatus.Phase = corev1.PodFailed
		} else {
			stepStatus.Phase = corev1.PodPhase(PhaseCompleted)
		}
		stepStatus.FinishedAt = status.State.Terminated.FinishedAt
		stepStatus.Reason = status.State.Terminated.Reason
	}

	return stepStatus
}

// ZBuilderSpec defines the desired state of ZBuilder
type ZBuilderSpec struct {
	// INSERT ADDITIONAL SPEC FIELDS - desired state of cluster
	// Important: Run "make" to regenerate code after modifying this file

	// Foo is an example field of ZBuilder. Edit zbuilder_types.go to remove/update
	// Foo string `json:"foo,omitempty"`
	Project    corev1.ObjectReference `json:"project,omitempty"`
	SubProject string                 `json:"subProject,omitempty"`
	Branch     string                 `json:"branch,omitempty"`
	CommitHash string                 `json:"commitHash,omitempty"`
	ZService   corev1.ObjectReference `json:"zservice,omitempty"`
	Creator    string                 `json:"creator,omitempty"`
	Build      bool                   `json:"build,omitempty"`
	Package    bool                   `json:"package,omitempty"`
	Deploy     bool                   `json:"deploy,omitempty"`
}

// ZBuilderStatus defines the observed state of ZBuilder
type ZBuilderStatus struct {
	// INSERT ADDITIONAL STATUS FIELD - define observed state of cluster
	// Important: Run "make" to regenerate code after modifying this file
	Steps      []StepStatus `json:"steps,omitempty"`
	CreatedAt  *metav1.Time `json:"createdAt,omitempty"`
	FinishedAt *metav1.Time `json:"finishedAt,omitempty"`
	Phase      Phase        `json:"phase,omitempty"`
	Message    string       `json:"message,omitempty"`
	Reason     string       `json:"reason,omitempty"`
	RunAt      string       `json:"RunAt,omitempty"`
}

func NewZbuilderStatus(podStatus *corev1.PodStatus) ZBuilderStatus {
	zbuilderStatus := ZBuilderStatus{}
	if podStatus == nil {
		return zbuilderStatus
	}

	for _, cs := range podStatus.ContainerStatuses {
		zbuilderStatus.Steps = append(zbuilderStatus.Steps, NewStepStatus(cs))
	}
	zbuilderStatus.CreatedAt = podStatus.StartTime
	switch podStatus.Phase {
	case corev1.PodPending:
		zbuilderStatus.Phase = PhasePending
	case corev1.PodRunning:
		zbuilderStatus.Phase = PhaseBuild
	}

	zbuilderStatus.Message = podStatus.Message
	zbuilderStatus.RunAt = podStatus.HostIP

	return zbuilderStatus
}

func (zbuilderStatus *ZBuilderStatus) SetPhase(phase Phase) {
	zbuilderStatus.Phase = phase
}

//+kubebuilder:object:root=true
//+kubebuilder:subresource:status

// ZBuilder is the Schema for the zbuilders API
type ZBuilder struct {
	metav1.TypeMeta   `json:",inline"`
	metav1.ObjectMeta `json:"metadata,omitempty"`

	Spec   ZBuilderSpec   `json:"spec,omitempty"`
	Status ZBuilderStatus `json:"status,omitempty"`
}

//+kubebuilder:object:root=true

// ZBuilderList contains a list of ZBuilder
type ZBuilderList struct {
	metav1.TypeMeta `json:",inline"`
	metav1.ListMeta `json:"metadata,omitempty"`
	Items           []ZBuilder `json:"items"`
}

func init() {
	SchemeBuilder.Register(&ZBuilder{}, &ZBuilderList{})
}

//返回workdir的volume mount
func (zb *ZBuilder) VolumeMountWorkDir() corev1.VolumeMount {
	return corev1.VolumeMount{
		Name:      "workdir",
		MountPath: "/workdir",
	}
}

//返回dockerconfig 的volumeMount
func (zb *ZBuilder) VolumeMountDockerConfig() corev1.VolumeMount {
	return corev1.VolumeMount{
		Name:      "dockerconfig",
		MountPath: "/root/.docker",
	}
}

//返回workdir的volume
func (zb *ZBuilder) VolumeWorkDir() corev1.Volume {
	return corev1.Volume{
		Name: "workdir",
		VolumeSource: corev1.VolumeSource{
			EmptyDir: &corev1.EmptyDirVolumeSource{},
		},
	}
}

func (zb *ZBuilder) VolumeDockerConfig(secret string) corev1.Volume {
	return corev1.Volume{
		Name: "dockerconfig",
		VolumeSource: corev1.VolumeSource{
			Secret: &corev1.SecretVolumeSource{
				SecretName: secret,
			},
		},
	}
}

//编译容器
func (zb *ZBuilder) Containers(template *ZBuildTemplateSpec, workdir string) []corev1.Container {
	containers := []corev1.Container{}
	for _, step := range template.Steps {
		volumeMounts := []corev1.VolumeMount{zb.VolumeMountWorkDir(), zb.VolumeMountDockerConfig()}
		volumeMounts = append(volumeMounts, step.VolumeMounts...)
		container := corev1.Container{
			Image:        step.Image,
			VolumeMounts: volumeMounts,
			Args:         step.Args,
			Command:      step.Commands,
			Env:          utils.MergeEnv(utils.MergeOverrideByFirst, zb.DefaultEnv(), step.Env),
			Name:         step.Name,
			WorkingDir:   workdir,
		}
		containers = append(containers, container)
	}
	return containers
}

//初始化容器，下载git文件
func (zb *ZBuilder) InitContainers(gitImage, gitRepo, dockerfile, dockerfilePath string) []corev1.Container {
	containers := []corev1.Container{
		//下载git仓库
		{
			Image:        gitImage,
			VolumeMounts: []corev1.VolumeMount{zb.VolumeMountWorkDir()},
			Command:      []string{"git"},
			Args:         []string{"clone", "-b", zb.Spec.Branch, gitRepo},
			WorkingDir:   "/workdir",
		},
		//写入Dockerfile
		{
			Image:        gitImage,
			VolumeMounts: []corev1.VolumeMount{zb.VolumeMountWorkDir()},
			Command:      []string{"echo"},
			Args:         []string{"-e", "${DOCKERFILE}", ">", dockerfilePath},
			WorkingDir:   "/workdir",
			Env: []corev1.EnvVar{
				{
					Name:  "DOCKERFILE",
					Value: dockerfile,
				},
			},
		},
	}
	return containers
}

//根据builder创建podspec
func (zb *ZBuilder) PodSpec(template *ZBuildTemplateSpec, gitImage string, gitRepo string, subProject *SubProject, registry *ZRegistry) corev1.PodSpec {
	vols := []corev1.Volume{zb.VolumeWorkDir(), zb.VolumeDockerConfig(registry.GetName())}
	vols = append(vols, subProject.BuildSpec.Volumes...)
	spec := corev1.PodSpec{
		InitContainers: zb.InitContainers(gitImage, gitRepo, subProject.BuildSpec.Dockerfile, zb.WorkPath(zb.RepoPath(gitRepo), subProject.Path, "Dockerfile")),
		Containers:     zb.Containers(template, zb.WorkPath(zb.RepoPath(gitRepo), subProject.Path)),
		// ImagePullSecrets: secrets,
		Volumes:       vols,
		RestartPolicy: corev1.RestartPolicyNever,
	}
	return spec
}

func (zb *ZBuilder) Pod(template *ZBuildTemplateSpec, gitImage string, gitRepo string, subProject *SubProject, registry *ZRegistry) corev1.Pod {
	spec := zb.PodSpec(template, gitImage, gitRepo, subProject, registry)
	pod := corev1.Pod{
		ObjectMeta: metav1.ObjectMeta{
			Name:      zb.GetName(),
			Namespace: zb.GetNamespace(),
			OwnerReferences: []metav1.OwnerReference{
				zb.Owner(),
			},
			Labels: zb.ResourceLabels(),
		},
		Spec: spec,
	}
	return pod
}

//根据repo获取project path
func (zb *ZBuilder) RepoPath(repo string) string {
	strs := strings.Split(repo, "/")
	if len(strs) < 1 {
		return ""
	}
	str := strs[len(strs)-1]
	strs = strings.Split(str, ".")
	if len(strs) < 1 {
		return str
	}
	return strs[0]
}

func (zb *ZBuilder) WorkPath(spath ...string) string {
	return path.Join("/workdir", path.Join(spath...))
}

//这里返回默认env，默认env提供一些基本信息
func (zb *ZBuilder) DefaultEnv() []corev1.EnvVar {
	return []corev1.EnvVar{
		{Name: "BRANCH", Value: zb.Spec.Branch},
		{Name: "COMMIT_HASH", Value: zb.Spec.CommitHash},
	}
}

//Owner
func (zb *ZBuilder) Owner() metav1.OwnerReference {
	bTrue := true
	return metav1.OwnerReference{
		APIVersion:         GroupVersion.Version,
		Kind:               "ZBuilder",
		Name:               zb.GetName(),
		BlockOwnerDeletion: &bTrue,
		Controller:         &bTrue,
		UID:                zb.GetUID(),
	}
}

//Labels
func (zb *ZBuilder) ResourceLabels() map[string]string {
	labels := make(map[string]string)
	labels["builders.zelda.io"] = zb.Name
	labels["projects.zelda.io"] = zb.Spec.Project.Name
	labels["users.zelda.io"] = zb.Spec.Creator
	labels["projects.zelda.io/branch"] = zb.Spec.Branch
	return labels
}
