/*
Copyright 2023 ycyxuehan.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package v1alpha1

import (
	"encoding/json"
	"fmt"
	"strings"

	"github.com/ycyxuehan/zelda/pkg/utils"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/types"
)

// EDIT THIS FILE!  THIS IS SCAFFOLDING FOR YOU TO OWN!
// NOTE: json tags are required.  Any new fields you add must have json tags for the fields to be serialized.

// ZRegistrySpec defines the desired state of ZRegistry
type ZRegistrySpec struct {
	// INSERT ADDITIONAL SPEC FIELDS - desired state of cluster
	// Important: Run "make" to regenerate code after modifying this file

	// Foo is an example field of ZRegistry. Edit zregistry_types.go to remove/update
	// Foo string `json:"foo,omitempty"`
	Server    string `json:"server,omitempty"`
	Namespace string `json:"namespace,omitempty"`
	Username  string `json:"username,omitempty"`
	Password  string `json:"password,omitempty"`
	Portal    string `json:"portal,omitempty"`
}

// ZRegistryStatus defines the observed state of ZRegistry
type ZRegistryStatus struct {
	// INSERT ADDITIONAL STATUS FIELD - define observed state of cluster
	// Important: Run "make" to regenerate code after modifying this file
}

//+kubebuilder:object:root=true
//+kubebuilder:subresource:status

// ZRegistry is the Schema for the zregistries API
type ZRegistry struct {
	metav1.TypeMeta   `json:",inline"`
	metav1.ObjectMeta `json:"metadata,omitempty"`

	Spec   ZRegistrySpec   `json:"spec,omitempty"`
	Status ZRegistryStatus `json:"status,omitempty"`
}

//+kubebuilder:object:root=true

// ZRegistryList contains a list of ZRegistry
type ZRegistryList struct {
	metav1.TypeMeta `json:",inline"`
	metav1.ListMeta `json:"metadata,omitempty"`
	Items           []ZRegistry `json:"items"`
}

func init() {
	SchemeBuilder.Register(&ZRegistry{}, &ZRegistryList{})
}

//返回imagetag的registry部分
func (zr *ZRegistry) FullImageTag(baseTag string) string {
	if zr.Spec.Server == "" {
		return baseTag
	}
	if zr.Spec.Namespace == "" {
		return strings.Join([]string{zr.Spec.Server, baseTag}, "/")
	}
	return strings.Join([]string{zr.Spec.Server, zr.Spec.Namespace, baseTag}, "/")
}

//返回登陆registry的secret名称，不需要登陆，返回空
func (zr *ZRegistry) ImagePullSecretName() string {
	if zr.Spec.Username == "" || zr.Spec.Password == "" {
		return ""
	}
	return zr.GetName()
}

//返回登陆registry的secret reference，不需要登陆，返回nil
func (zr *ZRegistry) ImagePullSecret() *corev1.LocalObjectReference {
	if zr.Spec.Username == "" || zr.Spec.Password == "" {
		return nil
	}
	return &corev1.LocalObjectReference{
		Name: zr.GetName(),
	}
}

//返回子资源secret的namespaced name
func (zr *ZRegistry) NamespacedNameSecret() types.NamespacedName {
	return types.NamespacedName{
		Namespace: zr.GetNamespace(),
		Name:      zr.GetName(),
	}
}

//返回子资源secret的name
func (zr *ZRegistry) SecreteName() string {
	return zr.GetName()
}

//
func (zr *ZRegistry) Owner() metav1.OwnerReference {
	bTrue := true
	return metav1.OwnerReference{
		Name:               zr.GetName(),
		APIVersion:         GroupVersion.Version,
		Kind:               "ZRegistry",
		UID:                zr.GetUID(),
		Controller:         &bTrue,
		BlockOwnerDeletion: &bTrue,
	}
}

func (zr *ZRegistry) ResourceLabels() map[string]string {
	labels := make(map[string]string)
	labels["registries.zelda.io"] = zr.GetName()
	return labels
}

//返回指定namespace的secret
func (zr *ZRegistry) NamespacedSecret(namespace string) *corev1.Secret {
	stringData := make(map[string]string)
	config := utils.NewDockerConfig(zr.Spec.Server, zr.Spec.Username, zr.Spec.Password)
	stringData["config.json"] = config.String()
	stringData[".dockerconfigjson"] = config.String()
	secret := corev1.Secret{
		ObjectMeta: metav1.ObjectMeta{
			Name:      zr.GetName(),
			Namespace: namespace,
			OwnerReferences: []metav1.OwnerReference{
				zr.Owner(),
			},
			Labels: zr.ResourceLabels(),
		},
		StringData: stringData,
	}
	return &secret
}

//返回registry所在namespace的secrete
func (zr *ZRegistry) Secret() *corev1.Secret {
	return zr.NamespacedSecret(zr.GetNamespace())
}

func (zr *ZRegistry) SecretReference() corev1.LocalObjectReference {
	return corev1.LocalObjectReference{
		Name: zr.GetName(),
	}
}

//返回zuser是否有变更
func (zr *ZRegistry) IsSpecChanged() (bool, error) {
	old := ZRegistry{}
	latestConfiguration, ok := zr.GetAnnotations()["kubectl.kubernetes.io/last-applied-configuration"]
	if !ok {
		return false, fmt.Errorf("not found annotation")
	}
	err := json.Unmarshal([]byte(latestConfiguration), &old)
	if err != nil { //无法获得旧配置
		return false, err
	}

	oldString, err := json.Marshal(&old.Spec)
	if err != nil {
		return false, err
	}
	newString, err := json.Marshal(&zr.Spec)
	if err != nil {
		return false, err
	}
	if string(oldString) != string(newString) {
		return true, nil
	}
	return false, nil
}
