/*
Copyright 2023 ycyxuehan.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package v1alpha1

import (
	"encoding/json"
	"fmt"
	"strings"

	"github.com/ycyxuehan/zelda/pkg/utils"
	appsv1 "k8s.io/api/apps/v1"
	batchv1 "k8s.io/api/batch/v1"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/types"
	"sigs.k8s.io/controller-runtime/pkg/client"
)

// EDIT THIS FILE!  THIS IS SCAFFOLDING FOR YOU TO OWN!
// NOTE: json tags are required.  Any new fields you add must have json tags for the fields to be serialized.

type ServiceSpec struct {
	Name  string               `json:"name,omitempty"`
	Ports []corev1.ServicePort `json:"ports,omitempty"`
	Type  corev1.ServiceType   `json:"type,omitempty"`
}

func (ss *ServiceSpec) ServiceSpec(selector map[string]string) corev1.ServiceSpec {
	spec := corev1.ServiceSpec{
		Ports:    ss.Ports,
		Selector: selector,
		Type:     ss.Type,
	}
	return spec
}

//Pod status
type PodStatus struct {
	Name    string          `json:"name,omitempty"`
	Phase   corev1.PodPhase `json:"phase,omitempty"`
	RunAt   string          `json:"runAt,omitempty"`
	Message string          `json:"message,omitempty"`
}

type PodStatusMap map[string]PodStatus

//Deployment status
type DeploymentStatus struct {
	AvailableReplicas   int32 `json:"availableReplicas,omitempty"`
	ReadyReplicas       int32 `json:"readyReplicas,omitempty"`
	Replicas            int32 `json:"replicas,omitempty"`
	UpdatedReplicas     int32 `json:"updatedReplicas,omitempty"`
	UnavailableReplicas int32 `json:"unavailableReplicas,omitempty"`
}

//Job status
type JobStatus struct {
	CompletionTime *metav1.Time `json:"completionTime,omitempty"`
	Succeeded      int32        `json:"succeeded,omitempty"`
	StartTime      *metav1.Time `json:"startTime,omitempty"`
	Active         int32        `json:"active,omitempty"`
	Failed         int32        `json:"failed,omitempty"`
}

type RunSpec struct {
	Template corev1.ObjectReference `json:"template,omitempty"`
	Volumes  []corev1.Volume        `json:"volumes,omitempty"`
}

// ZServiceSpec defines the desired state of ZService
type ZServiceSpec struct {
	// INSERT ADDITIONAL SPEC FIELDS - desired state of cluster
	// Important: Run "make" to regenerate code after modifying this file

	// Foo is an example field of ZService. Edit zservice_types.go to remove/update
	// Foo string `json:"foo,omitempty"`
	Env     []corev1.EnvVar `json:"env,omitempty"`
	Version string          `json:"version,omitempty"`
	// Volumes         []corev1.Volume        `json:"volumes,omitempty"`
	Services        []ServiceSpec          `json:"services,omitempty"`
	CustomHost      corev1.ObjectReference `json:"customHosts,omitempty"`
	NodeSelector    corev1.NodeSelector    `json:"nodeSelector,omitempty"`
	PodAffinity     corev1.PodAffinity     `json:"podAffinity,omitempty"`
	PodAntiAffinity corev1.PodAntiAffinity `json:"podAntiAffinity,omitempty"`
	Kind            string                 `json:"kind,omitempty"`
	Creator         corev1.ObjectReference `json:"creator,omitempty"`
	Project         corev1.ObjectReference `json:"project,omitempty"`
	SubProject      string                 `json:"subProject,omitempty"`
	RunSpec         RunSpec                `json:"runSpec,omitempty"`
}

// ZServiceStatus defines the observed state of ZService
type ZServiceStatus struct {
	// INSERT ADDITIONAL STATUS FIELD - define observed state of cluster
	// Important: Run "make" to regenerate code after modifying this file
	Branch           string           `json:"branch,omitempty"`
	Target           string           `json:"target,omitempty"`
	PodStatus        PodStatusMap     `json:"podStatus,omitempty"`
	DeploymentStatus DeploymentStatus `json:"deploymentStatus,omitempty"`
	JobStatus        JobStatus        `json:"jobStatus,omitempty"`
	UpdatedAt        *metav1.Time     `json:"updatedAt,omitempty"`
}

//+kubebuilder:object:root=true
//+kubebuilder:subresource:status

// ZService is the Schema for the zservices API
type ZService struct {
	metav1.TypeMeta   `json:",inline"`
	metav1.ObjectMeta `json:"metadata,omitempty"`

	Spec   ZServiceSpec   `json:"spec,omitempty"`
	Status ZServiceStatus `json:"status,omitempty"`
}

//+kubebuilder:object:root=true

// ZServiceList contains a list of ZService
type ZServiceList struct {
	metav1.TypeMeta `json:",inline"`
	metav1.ListMeta `json:"metadata,omitempty"`
	Items           []ZService `json:"items"`
}

func init() {
	SchemeBuilder.Register(&ZService{}, &ZServiceList{})
}

type NewClientObjectOption struct {
	ZRunTemplateSpec *ZRunTemplateSpec
	Registry         *ZRegistry
	UserSecrets      []corev1.LocalObjectReference
	HostAlias        []corev1.HostAlias
}

//返回子对象的namespaced name
func (zs *ZService) NamespacedNameResource() types.NamespacedName {
	return types.NamespacedName{
		Namespace: zs.GetNamespace(),
		Name:      zs.GetName(),
	}
}

//返回runTemplate的namespaced name
func (zs *ZService) NamespacedNameRunTemplate() types.NamespacedName {
	return types.NamespacedName{
		Namespace: zs.Spec.RunSpec.Template.Namespace,
		Name:      zs.Spec.RunSpec.Template.Name,
	}
}

//返回project的namespaced name
func (zs *ZService) NamespacedNameProject() types.NamespacedName {
	return types.NamespacedName{
		Namespace: zs.Spec.Project.Namespace,
		Name:      zs.Spec.Project.Name,
	}
}

//返回CustomHost的namespaced name
func (zs *ZService) NamespacedNameCustomHost() types.NamespacedName {
	return types.NamespacedName{
		Namespace: zs.Spec.CustomHost.Namespace,
		Name:      zs.Spec.CustomHost.Name,
	}
}

//返回user的namespaced name
func (zs *ZService) NamespacedNameUser() types.NamespacedName {
	return types.NamespacedName{
		Namespace: zs.Spec.Creator.Namespace,
		Name:      zs.Spec.Creator.Name,
	}
}

//返回zservice的owner reference
func (zs *ZService) Owner() metav1.OwnerReference {
	bTrue := true
	return metav1.OwnerReference{
		Name:               zs.GetName(),
		APIVersion:         GroupVersion.Version,
		Kind:               "ZService",
		UID:                zs.GetUID(),
		Controller:         &bTrue,
		BlockOwnerDeletion: &bTrue,
	}
}

//返回子资源的labels
func (zs *ZService) ResourceLabels() map[string]string {
	labels := make(map[string]string)
	labels["zservices.zelda.io"] = zs.GetName()
	labels["projects.zelda.io"] = zs.Spec.Project.Name
	labels["projects.selda.io/subproject"] = zs.Spec.SubProject
	return labels
}

//返回子资源的模板的labels,这里使用不一样的label，防止重复更新状态
func (zs *ZService) ResourceTemplateLabels() map[string]string {
	labels := make(map[string]string)
	labels["zservices.zelda.io"] = zs.GetName()
	// labels["projects.zelda.io"] = zs.Spec.Project.Name
	// labels["projects.selda.io/subproject"] = zs.Spec.SubProject
	return labels
}

//返回子资源selector
func (zs *ZService) Selector() *metav1.LabelSelector {
	return &metav1.LabelSelector{
		MatchLabels: zs.ResourceTemplateLabels(),
	}
}

//返回项目的imagetag
func (zs *ZService) ImageTagBase() string {
	return strings.Join([]string{zs.Spec.SubProject, zs.Spec.Version}, ":")
}

//更新模板中的image为项目的image
func (zs *ZService) UpdateContainerImage(image string, contaienrs ...corev1.Container) []corev1.Container {
	res := []corev1.Container{}
	for _, c := range contaienrs {
		if c.Image == "$PROJECT_IMAGE" {
			c.Image = image
		}
		res = append(res, c)
	}
	return res
}

//合并环境变量
func (zs *ZService) MergeContainerEnv(containers ...corev1.Container) []corev1.Container {
	res := []corev1.Container{}
	for _, c := range containers {
		envs := zs.Spec.Env
		envs = append(envs, c.Env...)
		mergedEnv := utils.MergeEnv(utils.MergeOverrideByFirst, envs)
		c.Env = mergedEnv
		res = append(res, c)
	}
	return res
}

//返回image tag
func (zs *ZService) ImageTag(registry *ZRegistry) string {
	image := zs.ImageTagBase()
	if registry != nil {
		image = registry.FullImageTag(image)
	}
	return image
}

//返回pod spec
func (zs *ZService) PodSpec(option *NewClientObjectOption) corev1.PodSpec {
	image := zs.ImageTag(option.Registry)
	secrets := option.UserSecrets
	if option.Registry != nil {
		secrets = append(secrets, corev1.LocalObjectReference{
			Name: option.Registry.ImagePullSecretName(),
		})
	}

	initContainers := zs.UpdateContainerImage(image, option.ZRunTemplateSpec.InitContainers...)
	containers := zs.UpdateContainerImage(image, option.ZRunTemplateSpec.Containers...)
	initContainers = zs.MergeContainerEnv(initContainers...)
	containers = zs.MergeContainerEnv(containers...)

	spec := corev1.PodSpec{
		InitContainers:   initContainers,
		Containers:       containers,
		ImagePullSecrets: secrets,
		Volumes:          zs.Spec.RunSpec.Volumes,
		HostAliases:      option.HostAlias,
	}
	return spec
}

//返回pod template
func (zs *ZService) PodTemplateSpec(option *NewClientObjectOption) corev1.PodTemplateSpec {
	spec := corev1.PodTemplateSpec{
		ObjectMeta: metav1.ObjectMeta{
			Labels: zs.ResourceTemplateLabels(),
		},
		Spec: zs.PodSpec(option),
	}
	return spec
}

//返回pod对象
func (zs *ZService) Pod(option *NewClientObjectOption) *corev1.Pod {
	pod := corev1.Pod{
		ObjectMeta: metav1.ObjectMeta{
			Name:      zs.GetName(),
			Namespace: zs.GetNamespace(),
			OwnerReferences: []metav1.OwnerReference{
				zs.Owner(),
			},
			Labels: zs.ResourceLabels(),
		},
		Spec: zs.PodSpec(option),
	}
	return &pod
}

//返回deployment spec
func (zs *ZService) DeploymentSpec(option *NewClientObjectOption) appsv1.DeploymentSpec {
	spec := appsv1.DeploymentSpec{
		Selector: zs.Selector(),
		Template: zs.PodTemplateSpec(option),
	}
	return spec
}

//返回deployment
func (zs *ZService) Deployment(option *NewClientObjectOption) *appsv1.Deployment {
	spec := zs.DeploymentSpec(option)
	deployment := appsv1.Deployment{
		ObjectMeta: metav1.ObjectMeta{
			Name:      zs.GetName(),
			Namespace: zs.GetNamespace(),
			OwnerReferences: []metav1.OwnerReference{
				zs.Owner(),
			},
			Labels: zs.ResourceLabels(),
		},
		Spec: spec,
	}

	return &deployment
}

//返回需要创建的service
func (zs *ZService) Services() []corev1.Service {
	services := []corev1.Service{}
	for _, svc := range zs.Spec.Services {
		objectMeta := metav1.ObjectMeta{
			Name:      svc.Name,
			Namespace: zs.GetNamespace(),
			OwnerReferences: []metav1.OwnerReference{
				zs.Owner(),
			},
			Labels: zs.ResourceLabels(),
		}
		service := corev1.Service{
			ObjectMeta: objectMeta,
			Spec:       svc.ServiceSpec(zs.ResourceTemplateLabels()),
		}
		services = append(services, service)
	}
	return services
}

//返回JobSpec
func (zs *ZService) JobSpec(option *NewClientObjectOption) batchv1.JobSpec {
	spec := batchv1.JobSpec{
		Selector: zs.Selector(),
		Template: zs.PodTemplateSpec(option),
	}
	return spec
}

//返回job
func (zs *ZService) Job(option *NewClientObjectOption) *batchv1.Job {
	job := batchv1.Job{
		ObjectMeta: metav1.ObjectMeta{
			Name:      zs.GetName(),
			Namespace: zs.GetNamespace(),
			OwnerReferences: []metav1.OwnerReference{
				zs.Owner(),
			},
			Labels: zs.ResourceLabels(),
		},
		Spec: zs.JobSpec(option),
	}
	return &job
}

//检查spec是否有变更
func (zs *ZService) IsSpecChanged() (bool, error) {
	oldZService := ZService{}
	latestConfiguration, ok := zs.GetAnnotations()["kubectl.kubernetes.io/last-applied-configuration"]
	if !ok {
		return false, fmt.Errorf("not found annotation")
	}
	err := json.Unmarshal([]byte(latestConfiguration), &oldZService)
	if err != nil { //无法获得旧配置
		return false, err
	}

	oldString, err := json.Marshal(&oldZService.Spec)
	if err != nil {
		return false, err
	}
	newString, err := json.Marshal(&zs.Spec)
	if err != nil {
		return false, err
	}
	if string(oldString) != string(newString) {
		return true, nil
	}
	return false, nil
}

//返回client object
func (zs *ZService) ClientObject(option *NewClientObjectOption) client.Object {
	switch zs.Spec.Kind {
	case "Deployment":
		return zs.Deployment(option)
	case "Job":
		return zs.Job(option)
	case "Pod":
		return zs.Pod(option)
	default:
		return nil

	}
}

//更新状态, 更新deployment
func (zs *ZService) UpdateStatusDeployment(name string, status *appsv1.DeploymentStatus) {
	if status == nil {
		return
	}
	zs.Status.Target = name
	zs.Status.DeploymentStatus = DeploymentStatus{
		AvailableReplicas:   status.AvailableReplicas,
		ReadyReplicas:       status.ReadyReplicas,
		Replicas:            status.Replicas,
		UpdatedReplicas:     status.UpdatedReplicas,
		UnavailableReplicas: status.UnavailableReplicas,
	}
}

//更新状态，更新job
func (zs *ZService) UpdateStatusJob(name string, status *batchv1.JobStatus) {
	if status == nil {
		return
	}
	zs.Status.Target = name
	zs.Status.JobStatus = JobStatus{
		CompletionTime: status.CompletionTime,
		Succeeded:      status.Succeeded,
		StartTime:      status.StartTime,
		Failed:         status.Failed,
		Active:         status.Active,
	}
}

//更新状态， 更新pod
func (zs *ZService) UpdateStatusPod(name string, status *corev1.PodStatus) {
	if status == nil {
		return
	}
	if zs.Spec.Kind == "Pod" {
		zs.Status.Target = name
	}
	podStatus := PodStatus{
		Name:    name,
		Phase:   status.Phase,
		RunAt:   status.HostIP,
		Message: status.Message,
	}
	zs.Status.PodStatus[name] = podStatus
}

//更新状态
func (zs *ZService) UpdateStatus(managedResource client.Object) error {
	switch zs.Spec.Kind {
	case "Deployment":
		dp, ok := managedResource.(*appsv1.Deployment)
		if !ok {
			return fmt.Errorf("object type error, it not a deployment")
		}
		zs.UpdateStatusDeployment(dp.GetName(), dp.Status.DeepCopy())
	case "Job":
		job, ok := managedResource.(*batchv1.Job)
		if !ok {
			return fmt.Errorf("object type error, it not a deployment")
		}
		zs.UpdateStatusJob(job.GetName(), job.Status.DeepCopy())
	case "Pod":
		pod, ok := managedResource.(*corev1.Pod)
		if !ok {
			return fmt.Errorf("object type error, it not a deployment")
		}
		zs.UpdateStatusPod(pod.GetName(), pod.Status.DeepCopy())
	default:
		return fmt.Errorf("unkown kind")
	}
	now := metav1.Now()
	zs.Status.UpdatedAt = &now
	return nil
}
