/*
Copyright 2023 ycyxuehan.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package v1alpha1

import (
	"encoding/json"
	"fmt"

	"github.com/ycyxuehan/zelda/pkg/utils"
	corev1 "k8s.io/api/core/v1"
	rbacv1 "k8s.io/api/rbac/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/types"
)

// EDIT THIS FILE!  THIS IS SCAFFOLDING FOR YOU TO OWN!
// NOTE: json tags are required.  Any new fields you add must have json tags for the fields to be serialized.

// ZUserSpec defines the desired state of ZUser
type ZUserSpec struct {
	// INSERT ADDITIONAL SPEC FIELDS - desired state of cluster
	// Important: Run "make" to regenerate code after modifying this file

	// Foo is an example field of ZUser. Edit zuser_types.go to remove/update
	// Foo string `json:"foo,omitempty"`
	Password    string                   `json:"password,omitempty"`
	Group       string                   `json:"group,omitempty"`
	NotifyMedia string                   `json:"notifyMedia,omitempty"` //user接收消息的媒介
	Secrets     []corev1.ObjectReference `json:"secrets,omitempty"`
}

// ZUserStatus defines the observed state of ZUser
type ZUserStatus struct {
	// INSERT ADDITIONAL STATUS FIELD - define observed state of cluster
	// Important: Run "make" to regenerate code after modifying this file
	CreatedAt *metav1.Time `json:"createdAt,omitempty"`
	Token     string       `json:"token,omitempty"`
}

//+kubebuilder:object:root=true
//+kubebuilder:subresource:status

// ZUser is the Schema for the zusers API
type ZUser struct {
	metav1.TypeMeta   `json:",inline"`
	metav1.ObjectMeta `json:"metadata,omitempty"`

	Spec   ZUserSpec   `json:"spec,omitempty"`
	Status ZUserStatus `json:"status,omitempty"`
}

//+kubebuilder:object:root=true

// ZUserList contains a list of ZUser
type ZUserList struct {
	metav1.TypeMeta `json:",inline"`
	metav1.ListMeta `json:"metadata,omitempty"`
	Items           []ZUser `json:"items"`
}

func init() {
	SchemeBuilder.Register(&ZUser{}, &ZUserList{})
}

func (zu *ZUser) Owner() metav1.OwnerReference {
	bTrue := true
	return metav1.OwnerReference{
		Name:               zu.GetName(),
		APIVersion:         GroupVersion.Version,
		Kind:               "ZUser",
		UID:                zu.GetUID(),
		Controller:         &bTrue,
		BlockOwnerDeletion: &bTrue,
	}
}

func (zu *ZUser) ResourceLabels() map[string]string {
	labels := make(map[string]string)
	labels["users.zelda.io"] = zu.GetName()
	labels["groups.zelda.io"] = zu.Spec.Group
	return labels
}

//返回zgroup的namespacedname
func (zu *ZUser) NamespacedNameGroup() types.NamespacedName {
	return types.NamespacedName{
		Namespace: zu.GetNamespace(),
		Name:      zu.Spec.Group,
	}
}

//返回zgroup的namespacedname
func (zu *ZUser) NamespacedNameServiceAccount() types.NamespacedName {
	return types.NamespacedName{
		Namespace: zu.GetNamespace(),
		Name:      zu.GetName(),
	}
}

//返回zgroup的namespacedname
func (zu *ZUser) NamespacedNameRoleBinding() types.NamespacedName {
	return types.NamespacedName{
		Namespace: zu.GetNamespace(),
		Name:      zu.GetName(),
	}
}

//返回zgroup的namespacedname
func (zu *ZUser) NamespacedNameClusterRoleBinding() types.NamespacedName {
	return types.NamespacedName{
		Name: zu.GetName(),
	}
}

//返回user对应的ServiceAccount
func (zu *ZUser) ServiceAccount(groupSecrets []corev1.ObjectReference) *corev1.ServiceAccount {
	secrets, _ := utils.MergeObjectReferences(utils.MergeOverrideByLatest, groupSecrets, zu.Spec.Secrets) //这里忽略错误，因为mergetype不可能出错
	sa := corev1.ServiceAccount{
		ObjectMeta: metav1.ObjectMeta{
			Name:      zu.GetName(),
			Namespace: zu.GetNamespace(),
			OwnerReferences: []metav1.OwnerReference{
				zu.Owner(),
			},
			Labels: zu.ResourceLabels(),
		},
		Secrets: secrets,
	}
	return &sa
}

//返回ServiceAccount的RoleBinding
func (zu *ZUser) RoleBinding() *rbacv1.RoleBinding {
	roleBinding := rbacv1.RoleBinding{
		ObjectMeta: metav1.ObjectMeta{
			Name:      zu.GetName(),
			Namespace: zu.GetNamespace(),
			OwnerReferences: []metav1.OwnerReference{
				zu.Owner(),
			},
			Labels: zu.ResourceLabels(),
		},
		Subjects: []rbacv1.Subject{
			{
				Kind:      rbacv1.ServiceAccountKind,
				Namespace: zu.GetNamespace(),
				Name:      zu.GetName(),
			},
		},
		RoleRef: rbacv1.RoleRef{
			APIGroup: rbacv1.GroupName,
			Kind:     "Role",
			Name:     zu.Spec.Group,
		},
	}
	return &roleBinding
}

//ServiceAccount的ClusterRoleBinding
func (zu *ZUser) ClusterRoleBinding() *rbacv1.ClusterRoleBinding {
	roleBinding := rbacv1.ClusterRoleBinding{
		ObjectMeta: metav1.ObjectMeta{
			Name: zu.GetName(),
			OwnerReferences: []metav1.OwnerReference{
				zu.Owner(),
			},
			Labels: zu.ResourceLabels(),
		},
		Subjects: []rbacv1.Subject{
			{
				Kind:      rbacv1.ServiceAccountKind,
				Namespace: zu.GetNamespace(),
				Name:      zu.GetName(),
			},
		},
		RoleRef: rbacv1.RoleRef{
			APIGroup: rbacv1.GroupName,
			Kind:     "ClusterRole",
			Name:     zu.Spec.Group,
		},
	}
	return &roleBinding
}

//返回zuser是否有变更
func (zu *ZUser) IsSpecChanged() (bool, error) {
	old := ZUser{}
	latestConfiguration, ok := zu.GetAnnotations()["kubectl.kubernetes.io/last-applied-configuration"]
	if !ok {
		return false, fmt.Errorf("not found annotation")
	}
	err := json.Unmarshal([]byte(latestConfiguration), &old)
	if err != nil { //无法获得旧配置
		return false, err
	}

	oldString, err := json.Marshal(&old.Spec)
	if err != nil {
		return false, err
	}
	newString, err := json.Marshal(&zu.Spec)
	if err != nil {
		return false, err
	}
	if string(oldString) != string(newString) {
		return true, nil
	}
	return false, nil
}
