/*
Copyright 2023 ycyxuehan.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package controller

import (
	"context"

	rbacv1 "k8s.io/api/rbac/v1"
	"k8s.io/apimachinery/pkg/runtime"
	ctrl "sigs.k8s.io/controller-runtime"
	"sigs.k8s.io/controller-runtime/pkg/client"
	"sigs.k8s.io/controller-runtime/pkg/log"

	zeldaiov1alpha1 "github.com/ycyxuehan/zelda/api/v1alpha1"
)

// ZGroupReconciler reconciles a ZGroup object
type ZGroupReconciler struct {
	client.Client
	Scheme *runtime.Scheme
}

//+kubebuilder:rbac:groups=zelda.io,resources=zgroups,verbs=get;list;watch;create;update;patch;delete
//+kubebuilder:rbac:groups=zelda.io,resources=zgroups/status,verbs=get;update;patch
//+kubebuilder:rbac:groups=zelda.io,resources=zgroups/finalizers,verbs=update
//+kubebuilder:rbac:groups=rbac.authorization.k8s.io/v1,resources=roles,verbs=get;list;watch;create;update;patch;delete
//+kubebuilder:rbac:groups=rbac.authorization.k8s.io/v1,resources=roles/status,verbs=get;update;patch
//+kubebuilder:rbac:groups=rbac.authorization.k8s.io/v1,resources=roles/finalizers,verbs=update
//+kubebuilder:rbac:groups=rbac.authorization.k8s.io/v1,resources=clusterroles,verbs=get;list;watch;create;update;patch;delete
//+kubebuilder:rbac:groups=rbac.authorization.k8s.io/v1,resources=clusterroles/status,verbs=get;update;patch
//+kubebuilder:rbac:groups=rbac.authorization.k8s.io/v1,resources=clusterroles/finalizers,verbs=update



// Reconcile is part of the main kubernetes reconciliation loop which aims to
// move the current state of the cluster closer to the desired state.
// TODO(user): Modify the Reconcile function to compare the state specified by
// the ZGroup object against the actual cluster state, and then
// perform operations to make the cluster state reflect the state specified by
// the user.
//
// For more details, check Reconcile and its Result here:
// - https://pkg.go.dev/sigs.k8s.io/controller-runtime@v0.10.0/pkg/reconcile
func (r *ZGroupReconciler) Reconcile(ctx context.Context, req ctrl.Request) (ctrl.Result, error) {
	_ = log.FromContext(ctx)

	// TODO(user): your logic here
	zgroup := zeldaiov1alpha1.ZGroup{}
	err := r.Get(ctx, req.NamespacedName, &zgroup)
	if err != nil {
		return ctrl.Result{}, err
	}
	if zgroup.GetDeletionTimestamp() == nil || zgroup.GetDeletionTimestamp().IsZero() {
		err = r.Process(ctx, &zgroup)
		return ctrl.Result{}, err
	}
	return ctrl.Result{}, nil
}

// SetupWithManager sets up the controller with the Manager.
func (r *ZGroupReconciler) SetupWithManager(mgr ctrl.Manager) error {
	return ctrl.NewControllerManagedBy(mgr).
		For(&zeldaiov1alpha1.ZGroup{}).
		Complete(r)
}

//处理zgroup的变更
//zgroup会根据配置创建对应的 role， cluster role
func (r *ZGroupReconciler) Process(ctx context.Context, zgroup *zeldaiov1alpha1.ZGroup) error {
	//空group，暂不处理。group暂时没有status需要处理
	if zgroup.IsEmpty() {
		return nil
	}
	changed, _ := zgroup.IsSpecChanged()
	//处理role
	err := r.ProcessRole(ctx, zgroup, changed)
	if err != nil {
		//打印日志，继续处理cluster role
		log.Log.Error(err, "process role")
	}
	err = r.ProcessClusterRole(ctx, zgroup, changed)
	return err
}

//处理Role
func (r *ZGroupReconciler) ProcessRole(ctx context.Context, zgroup *zeldaiov1alpha1.ZGroup, changed bool) error {
	if len(zgroup.Spec.Rulers) == 0 {
		return nil
	}
	role := rbacv1.Role{}
	err := r.Get(ctx, zgroup.NamespacedNameRole(), &role)
	if !changed && err == nil {
		return nil
	}
	role = *zgroup.Role()
	if err != nil {
		err = r.Create(ctx, &role, &client.CreateOptions{})
		return err
	}
	err = r.Update(ctx, &role, &client.UpdateOptions{})
	return err
}

//处理cluster role
func (r *ZGroupReconciler) ProcessClusterRole(ctx context.Context, zgroup *zeldaiov1alpha1.ZGroup, changed bool) error {
	if len(zgroup.Spec.Rulers) == 0 {
		return nil
	}
	role := rbacv1.ClusterRole{}
	err := r.Get(ctx, zgroup.NamespacedNameRole(), &role)
	if !changed && err == nil {
		return nil
	}
	role = *zgroup.ClusterRole()
	if err != nil {
		err = r.Create(ctx, &role, &client.CreateOptions{})
		return err
	}
	err = r.Update(ctx, &role, &client.UpdateOptions{})
	return err
}
